/* bsgs.c -- (C) Geoffrey Reynolds, April 2006.

   Implementation of a baby step giant step algorithm for finding all
   n in the range nmin <= n <= nmax satisfying b^n=d_i (mod p) where b
   and each d_i are relatively prime to p.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include "sr5sieve.h"
#include "arithmetic.h"
#include "hashtable.h"

/*
  Giant step baby step algorithm, from Wikipaedea:

  input: A cyclic group G of order n, having a generator a and an element b.

  Output: A value x satisfying a^x = b (mod n).

  1. m <-- ceiling(sqrt(n))
  2. For all j where 0 <= j < m:
     1. Compute a^j mod n and store the pair (j, a^j) in a table.
  3. Compute a^(-m).
  4. g <-- b.
  5. For i = 0 to (m - 1):
     1. Check if g is the second component (a^j) of any pair in the table.
     2. If so, return im + j.
     3. If not, g <-- ga^(-m) mod n.
*/

static uint32_t m;    /* Number of baby steps */
static uint32_t M;    /* Number of giant steps */

static uint32_t sieve_low;
static uint32_t sieve_range;

static uint64_t *bd64, *ck64, *D64;

static uint32_t baby_steps64(uint64_t b, uint64_t bj0, uint64_t p)
{
  uint64_t bj;
  uint32_t j;

  clear_hashtable();
  pre2_mulmod64_init(b);
  for (j = 0, bj = bj0; j < m; j++)
  {
    insert64(j, bj);
    bj = pre2_mulmod64(bj,b,p);
    if (bj == bj0)
    {
      pre2_mulmod64_fini();
      return j+1;
    }
  }

  pre2_mulmod64_fini();
  return 0;
}

static uint64_t setup64(uint64_t p)
{
  uint64_t inv_b, bm;
  uint32_t i;

  /* Precompute -c/k (mod p) for each sequence k*b^n+c.
   */
  for (i = 0; i < seq_count; i++)
    ck64[i] = invmod32_64(SEQ[i].k,SEQ[i].c,p);

  /* Precompute 1/b^d (mod p) for 0 <= d <= Q.
  */
  /* bd64[0] = 1; */ /* initialised in init_bsgs() */
  inv_b = invmod32_64(BASE,-1,p);
  pre2_mulmod64_init(inv_b);
  for (i = 1, bm = inv_b; i < subseq_Q; i++)
  {
    bd64[i] = bm;
    bm = pre2_mulmod64(bm,inv_b,p);
  }
  pre2_mulmod64_fini();
  /* bm now contains 1/b^Q. */

  /* Compute -c/(k*b^d) (mod p) for each subsequence. */
  for (i = 0; i < subseq_count; i++)
    D64[i] = mulmod64(ck64[SUBSEQ[i].seq],bd64[SUBSEQ[i].d],p);

  return bm;
}

static void bsgs64(uint64_t p)
{
  uint32_t i, j, k;
  uint64_t b, bm;

#if USE_ASM && defined(__ppc64__)
  getMagic(p);
#endif

  mod64_init(p);

  /* bm <- 1/b^Q (mod p). */
  bm = setup64(p);

  /* b <- base^Q (mod p) */
  b = powmod64(BASE,subseq_Q,p);

  /* Baby steps. */
  if ((i = baby_steps64(b,powmod64(b,sieve_low,p),p)) > 0)
  {
    /* i is the order of b (mod p). This is all the information we need to
       determine every solution for this p, no giant steps are needed.
    */
    for (k = 0; k < subseq_count; k++)
      for (j = lookup64(D64[k]); j < sieve_range; j += i)
        eliminate_term(k,sieve_low+j,p);
    mod64_fini();
    return;
  }

  /* First giant step. */
  for (k = 0; k < subseq_count; k++)
    if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
      eliminate_term(k,sieve_low+j,p);

  /* Remaining giant steps. */
  bm = powmod64(bm,m,p); /* bm <- 1/b^m (mod p) */
  pre2_mulmod64_init(bm);
  for (i = 1; i < M; i++)
    for (k = 0; k < subseq_count; k++)
    {
      D64[k] = pre2_mulmod64(D64[k],bm,p);
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(k,sieve_low+i*m+j,p);
    }
  pre2_mulmod64_fini();
  mod64_fini();
}

static void init_bsgs(void)
{
  uint32_t r = n_max/subseq_Q-n_min/subseq_Q+1, s = subseq_count;

  /*
    In the worst case we will do do one table insertion and one mulmod
    for m baby steps, then s table lookups and s mulmods for M giant
    steps. The average case depends on how many solutions are found
    and how early in the loop they are found, which I don't know how
    to analyse. However for the worst case we just want to minimise
    m + s*M subject to m*M >= r, which is when m = sqrt(s*r).
  */
  M = MAX(1,sqrt((double)r/s));
  m = MIN(r,ceil((double)r/M));

  if (m > HASH_MAX_ELTS)
  {
    M = (double)r/HASH_MAX_ELTS;
    m = ceil((double)r/M);
  }

  sieve_low = n_min / subseq_Q;
  sieve_range = m*M;

  assert(sieve_low <= n_min/subseq_Q);
  assert(n_max/subseq_Q < sieve_low+sieve_range);

  ck64 = xmalloc(seq_count*sizeof(uint64_t));
  bd64 = xmalloc(subseq_Q*sizeof(uint64_t));
  bd64[0] = UINT64_C(1);
  D64 = xmalloc(subseq_count*sizeof(uint64_t));
  init_hashtable(m);
}

static void fini_bsgs(void)
{
  fini_hashtable();
  free(D64);
  free(bd64);
  free(ck64);
}

void sieve(void)
{
  init_bsgs();
  init_prime_sieve(p_max);
  start_srsieve();
  prime_sieve(p_min,p_max,bsgs64);
  finish_srsieve("range is complete",p_max);
  fini_prime_sieve();
  fini_bsgs();
}
