/* subseq.c -- (C) Geoffrey Reynolds, June 2006.

   Base b sequences in n=qm+r where n0 < n < n1 are represented as a number
   of base b^Q subsequences in m where n0/Q < m < n1/Q.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include "sr5sieve.h"
#include "bitmap.h"
#include "arithmetic.h"

subseq_t *SUBSEQ = NULL;
uint32_t subseq_count = 0;
uint32_t subseq_Q = 0;

static int subseq_clear_m(uint32_t subseq, uint32_t m)
{
  assert(subseq < subseq_count);

  if (m < SUBSEQ[subseq].m_low || m > SUBSEQ[subseq].m_high)
    return 0;

  return clear_bit(SUBSEQ[subseq].M, m - SUBSEQ[subseq].m_low);
}

static int subseq_set_m(subseq_t *SUBSEQ, uint32_t subseq, uint32_t m)
{
  assert (m >= SUBSEQ[subseq].m_low && m <= SUBSEQ[subseq].m_high);

  return set_bit(SUBSEQ[subseq].M, m - SUBSEQ[subseq].m_low);
}

void make_subseqs(void)
{
  uint32_t i, j, r, n, Q, s, s_count;
  subseq_t *tmp;

#ifndef NDEBUG
  uint32_t t_count = 0, old_count = 0;
#endif

  Q = find_best_Q(LIMIT_BASE,&s_count);

  tmp = xmalloc(s_count*sizeof(subseq_t));

  for (i = 0, s = 0; i < seq_count; i++)
  {
    uint32_t nmin[Q], nmax[Q], subseq[Q];

    for (j = 0; j < Q; j++)
      nmin[j] = UINT32_MAX, nmax[j] = 0;

    for (j = 0; j < SEQ[i].ncount; j++)
    {
      n = SEQ[i].N[j];
      r = n % Q;
      nmin[r] = MIN(nmin[r],n);
      nmax[r] = MAX(nmax[r],n);
    }

    for (r = 0; r < Q; r++)
      if (nmin[r] != UINT32_MAX)
      {
        tmp[s].seq = i;
        tmp[s].d = r;
        tmp[s].m_low = nmin[r]/Q;
        tmp[s].m_high = nmax[r]/Q;
        tmp[s].M = make_bitmap(tmp[s].m_high-tmp[s].m_low+1);
        tmp[s].mcount = 0;
        subseq[r] = s++;
      }

    assert (s <= s_count);

    for (j = 0; j < SEQ[i].ncount; j++)
    {
      n = SEQ[i].N[j];
      r = n % Q;
      subseq_set_m(tmp,subseq[r],n/Q);
      tmp[subseq[r]].mcount++;
    }

    free(SEQ[i].N);
    SEQ[i].nsize = 0;
    SEQ[i].type = seq_bitmap;
#ifndef NDEBUG
    old_count += SEQ[i].ncount;
#endif
  }

  assert (s == s_count);

  if (Q > 1)
    printf("Split %"PRIu32" base %"PRIu32" sequence%s into %"PRIu32
           " base %"PRIu32"^%"PRIu32" subsequence%s.\n", seq_count,
           BASE, plural(seq_count), s_count, BASE, Q, plural(s_count));

  subseq_Q = Q;
  SUBSEQ = tmp;
  subseq_count = s_count;

#ifndef NDEBUG
  for (i = 0, t_count = 0; i < subseq_count; i++)
    t_count += SUBSEQ[i].mcount;
#endif

  assert(t_count == old_count);
}

static void report_factor(uint32_t subseq, uint32_t n, uint64_t p)
{
  uint32_t k = SEQ[SUBSEQ[subseq].seq].k;
  int32_t c = SEQ[SUBSEQ[subseq].seq].c;

#if CHECK_FACTORS
  if (!is_factor(k,c,n,p))
    error("%"PRIu64" DOES NOT DIVIDE %s.",p,kbnc_str(k,n,c));
#endif

  factor_count++;
  SUBSEQ[subseq].mcount--;
  save_factor(k,c,n,p);
  printf("%"PRIu64" | %s\n",p,kbnc_str(k,n,c));
}

#if REPORT_DUPLICATES
static void report_duplicate_factor(uint32_t seq, uint32_t n, uint64_t p)
{
  uint32_t k = SEQ[seq].k;
  int32_t c = SEQ[seq].c;

#if CHECK_FACTORS
  if (!is_factor(k,c,n,p))
    error("%"PRIu64" DOES NOT DIVIDE %s (duplicate).",p,kbnc_str(k,n,c));
#endif

  if (n <= n_max)
    printf("%"PRIu64" | %s (duplicate)\n",p,kbnc_str(k,n,c));
}
#endif /* REPORT_DUPLICATES */

/* eliminate a single term n=Qm+d for this subsequence.
*/
void eliminate_term(uint32_t subseq, uint32_t m, uint64_t p)
{
  assert(subseq < subseq_count);

  if (subseq_clear_m(subseq,m))
    report_factor(subseq,m*subseq_Q+SUBSEQ[subseq].d,p);
#if REPORT_DUPLICATES
  else
    report_duplicate_factor(SUBSEQ[subseq].seq,m*subseq_Q+SUBSEQ[subseq].d,p);
#endif /* REPORT_DUPLICATES */
}
